/*
 * locate.c - client localization application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>

/*
TODO
- err handling for malloc (operator, country, ms/t)
*/

/* *INDENT-OFF* */

lc_char PROGMEM lc_Locate[]={
        LC_EN("Locate")
        LC_CZ("Lokalizuj")
        LC_END
};

lc_char PROGMEM lc_Self[]={
        LC_EN("My Location")
        LC_CZ("Moje poloha")
        LC_END
};

lc_char PROGMEM lc_Where_is[]={
        LC_EN("Where is?")
        LC_CZ("Kde je?")
        LC_END
};

lc_char PROGMEM lc_Sending_SMS[]={
        LC_EN("Sending Message")
        LC_CZ("Posilam zpravu")
        LC_END
};

lc_char PROGMEM lc_SMS_Send_OK[]={
        LC_EN("Message Sent")
        LC_CZ("Zprava odeslana")
        LC_END
};

lc_char PROGMEM lc_SMS_Send_Not_OK[]={
        LC_EN("Sending was not successful")
        LC_CZ("Zpravu se nepodarilo odeslat")
        LC_END
};

lc_char PROGMEM lc_No_Info[]={
        LC_EN("No Location Info Available")
        LC_CZ("Lokalizace nedostupna")
        LC_END
};

lc_char PROGMEM lc_Err_NO_DIR[]={
        LC_EN("Installation error")
        LC_CZ("Spatna instalace aplikace")
        LC_END
};

lc_char PROGMEM lc_Err_NO_LIST[]={
        LC_EN("No cell list for unknown operator: ")
        LC_CZ("Neni seznam bunek pro neznameho operatora: ")
        LC_END
};

lc_char PROGMEM lc_Err_UNKNOWN_CELL[]={
        LC_EN("Unknown cell: ")
        LC_CZ("Neznama bunka: ")
        LC_END
};

lc_char PROGMEM lc_Err_BROKEN_FILE[]={
        LC_EN("Broken cell list")
        LC_CZ("Poskozeny seznam bunek")
        LC_END
};

/* *INDENT-ON* */

u8 PROGMEM t_Position_mime[] = "locinfo";

u8 PROGMEM t_cells[] = { "/cells" };

#define TAG_CELL_TEXT   	0x01
#define TAG_CELLS_COUNTRY       0x02
#define TAG_CELLS_OPERATOR      0x03
#define TAG_CELLS_DATE          0x04

#define CELLS_NO_DIR		0x00
#define CELLS_NO_LIST		0x01
#define CELLS_UNKNOWN_CELL	0x02
#define CELLS_BROKEN_FILE	0x03

void err_to_buf (u8 err, u8 * buf, u8 * res)
{
  u8 i;
  u8 len;
  u8 *e;

  switch (err)
  {
    case CELLS_NO_DIR:
      e = locale (lc_Err_NO_DIR);
      break;
    case CELLS_NO_LIST:
      e = locale (lc_Err_NO_LIST);
      break;
    case CELLS_UNKNOWN_CELL:
      e = locale (lc_Err_UNKNOWN_CELL);
      break;
    case CELLS_BROKEN_FILE:
      e = locale (lc_Err_BROKEN_FILE);
      break;
  }

  len = strlen (e) + 1;
  memcpy (buf, e, len);

  if (err == CELLS_UNKNOWN_CELL)
  {
    len--;
    buf[len++] = ((res[5] & 0xF0) >> 4) + '0';
    buf[len++] = (res[5] & 0x0F) + '0';
    buf[len++] = ((res[6] & 0xF0) >> 4) + '0';
    buf[len++] = (res[6] & 0x0F) + '0';
    buf[len++] = '\0';
  }
  else if (err == CELLS_NO_LIST)
  {
    len--;
    buf[len++] = (res[0] & 0x0F) + '0';
    buf[len++] = ((res[0] & 0xF0) >> 4) + '0';
    buf[len++] = (res[1] & 0x0F) + '0';
    buf[len++] = (res[2] & 0x0F) + '0';
    buf[len++] = ((res[2] & 0xF0) >> 4) + '0';
    buf[len++] = '\0';
  }
}

void seek_cell (u8 * res, u8 * buf1)
{
  u8 buf2[7];
  MCFILE *f = NULL;
  u16 cid, next_val, lac, tmp_lac;
  b32 b;
  b32 file_pos;
  u8 len, i, j;
  u8 *operator = NULL;
  u8 *country = NULL;
  u8 *tmp;

  /*
     T_LOCATION_INFO:
     0,1,2 MCC, MNC
     3,4 LAC
     5,6 CID
   */

  buf2[0] = (res[0] & 0x0F) + '0';
  buf2[1] = ((res[0] & 0xF0) >> 4) + '0';
  buf2[2] = (res[0 + 1] & 0x0F) + '0';
  buf2[3] = (res[0 + 2] & 0x0F) + '0';
  buf2[4] = ((res[0 + 2] & 0xF0) >> 4) + '0';
  buf2[5] = '\0';

  b = cd (t_cells);
  if (b < 0)
  {
    dbsp ("no /cells\n");
    err_to_buf (CELLS_NO_DIR, buf1, res);
    goto X_ERR;
  }

  b = mc_find (0, buf2, FS_FIND_STRING);
  if (b < 0)
  {
    dbsp ("no cell list\n");
    err_to_buf (CELLS_NO_LIST, buf1, res);
    goto X_ERR;
  }

  cid = res[6] | (res[5] << 8);
  lac = res[4] | (res[3] << 8);
  dbsp ("XCID ");
  dbih (cid);
  dbc (' ');
  dbsp ("XLAC ");
  dbih (lac);
  dbc ('\n');

  f = mc_open (b, FILE_READ);

  mc_read (f, 0, buf2, 1);
  mc_read (f, 0, buf1, buf2[0] + 1);

  i = get_tag (buf1, TAG_CELLS_OPERATOR);
  if (i != 0)
  {
    operator = malloc (buf1[i + 1] + 1);

    memcpy (operator, &buf1[i + 2], buf1[i + 1]);
    operator[buf1[i + 1]] = '\0';
  }

  i = get_tag (buf1, TAG_CELLS_COUNTRY);
  if (i != 0)
  {
    country = malloc (buf1[i + 1] + 1);
    memcpy (country, &buf1[i + 2], buf1[i + 1]);
    country[buf1[i + 1]] = '\0';
  }

  file_pos = buf2[0] + 1;	//skip head

  dbsp (">seek_cid\n");

  while (1)
  {
    mc_read (f, file_pos, buf2, 3);
    next_val = buf2[0] | (buf2[1] << 8);
    len = buf2[2];

    dbsp ("cell stamp ");
    dbih (next_val);
    dbc (' ');
    dbch (len);
    dbc ('\n');

    if (next_val == 0 || cid < next_val)
      break;
    file_pos += len * 8 + 3;
  }

  file_pos += 3;

  for (i = 0; i < len; i++)
  {
    mc_read (f, file_pos + i * 8, buf2, 4);
    tmp_lac = buf2[2] | (buf2[3] << 8);
    dbsp ("LAC ");
    dbih (tmp_lac);
    dbc (' ');
    next_val = buf2[0] | (buf2[1] << 8);
    dbsp ("CID ");
    dbih (next_val);
    dbc (' ');
    dbch (i);
    dbc ('\n');

    if (cid == next_val)
    {
      next_val = 0;

      if (i + 1 < len)
      {
	mc_read (f, file_pos + (i + 1) * 8, buf2, 2);
	next_val = buf2[0] | (buf2[1] << 8);
      }

      if (cid != next_val || lac == tmp_lac)	// either both cid & lac valid or last cid (then ignore lac)
      {
	mc_read (f, file_pos + i * 8 + 4, buf2, 4);
	file_pos =
	  ((u32) buf2[0]) | ((u32) buf2[1] << 8) | ((u32) buf2[2] << 16) |
	  ((u32) buf2[3] << 24);
	dbsp ("cid found ");
	dbih (file_pos >> 16);
	dbih (file_pos);
	dbc ('\n');
	goto X_FOUND;
      }
    }
  }

  dbsp ("cid not found\n");
  err_to_buf (CELLS_UNKNOWN_CELL, buf1, res);
  goto X_ERR;

X_FOUND:
  dbsp ("<seek_cid\n");

  len = 0;
  tmp = buf1;

  if (operator)
  {
    i = strlen (operator);
    operator[i] = '\n';
    i++;
    memcpy (tmp, operator, i);
    tmp += i;
    len += i;
  }

  if (country)
  {
    i = strlen (country);
    country[i] = '\n';
    i++;
    memcpy (tmp, country, i);
    tmp += i;
    len += i;
  }

  /*
     tmp=sprints(tmp, t_LAC); len+=sizeof(t_LAC)-1;
     tmp=sprintch(tmp, res[3]); len+=2;
     tmp=sprintch(tmp, lac); len+=2;
     tmp=sprintc(tmp, ' '); len+=1;
   */

  mc_read (f, file_pos, tmp, 1);
  i = tmp[0];
  mc_read (f, file_pos, tmp, i + 1);

  i = get_tag (tmp, TAG_CELL_TEXT);
  if (i != 0)
  {
    u8 l1;

    dbsp ("tag found ");
    dbch (i);
    dbc ('\n');
    l1 = tmp[i + 1];
    for (j = 0; j < l1; j++)
      tmp[j] = tmp[i + j + 2];
    l1 = dcs_7b_len (l1);
    len += dcs_78 (tmp, l1, DCS_7_TO_8);
    buf1[len] = '\0';
  }
  else
    err_to_buf (CELLS_BROKEN_FILE, buf1, res);

X_ERR:
  mc_close (f);

  free (operator);
  free (country);
}

u8 locate_whereis (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *tmp = NULL;
    u8 *ms = NULL;
    u8 *res;
    u8 *q;
    u8 l, j, i;

    res = msisdn (NULL);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;

    ms = msisdncpy (res, MSISDN_ADN, MEM_R);
    tmp = malloc (20);
    q = create_head_q (tmp, t_Position_mime, ms, MSISDN_ADN);
    res =
      send_sms (tmp, q - tmp, ms, MSISDN_ADN, tsms_dcs (), tsms_pid (),
		locale (lc_Sending_SMS), NULL);

    i = get_tag (res, T_RESULT);
    if (i != 0)
    {
      i += 2;
      if (res[i] == 0x00)
	i = display_text (locale (lc_SMS_Send_OK), NULL);
      else
	i = display_text (locale (lc_SMS_Send_Not_OK), NULL);
    }
    else
      i = APP_END;

    free (tmp);
    free (ms);
    return i;
  }

  return APP_OK;
}

u8 locate_self (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res;
    u8 i;
    u8 *buf1 = buf_B ();

    res = provide_local_info ();
    i = get_tag (res, T_LOCATION_INFO);
    if (i != 0)
    {
      seek_cell (res + i + 2, buf1);	//skip tag, len
      return display_text (buf1, NULL);
    }
    return display_text (locale (lc_No_Info), NULL);
  }
  return APP_OK;
}

void action_answer (u8 * s)
{
  u8 *t = skip_head (s);
  u8 j;
  u8 *res;
  u8 i, l;
  u16 cid;
  u8 *buf1 = buf_B ();
  u8 buf2[7];
  b32 file_pos;
  u8 *ms = NULL;

  res = t + 1;			//skip terminal profile

  j = get_tag (res, T_LOCATION_INFO);
  if (j == 0)
    return;
  j += 2;			//skip tag, len

  memcpy (buf2, res + j, 7);

  t = msisdncpy (tpdu_seek (s, T_SMS_AA), MSISDN_SMS, MEM_R);
  ms = get_adn (t, MSISDN_SMS);

#ifdef DEBUG
  {
    u8 xi = strlen (ms);

    dbsp ("MSX: ");
    for (i = 0; i < xi; i++)
      dbc (ms[i]);
    dbc ('\n');
  }
#endif

  free (t);
  more_time ();
  seek_cell (buf2, buf1);
  display_text (buf1, ms);

X_ERR:
  free (ms);
}

SNodeP locate_n = { NULL, NULL };
SNodeP locate_n_self = { lc_Self, locate_self };
SNodeP locate_n_whereis = { lc_Where_is, locate_whereis };

SEdgeP locate_edges_p[] = {
  {&locate_n, &locate_n_whereis}
  ,
  {&locate_n, &locate_n_self}
  ,
  NULL
};

void locate (Menu_selection_data * x)
{
  SCtx *c;
  u8 i;

  i = mc_detect ();
  if (i == MC_NOT_PRESENT)
  {
    perror (ERR_MC_NOT_PRESENT);
    return;
  }
  else if (i != MC_OK)
  {
    perror (ERR_MC_NOT_READY);
    return;
  }

  c = spider_init (c);
  if (c == NULL)
    return;

  c->n = &locate_n;
  c->eP = &locate_edges_p;

  spider (c);
}

void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_SMS_A:
      stk_thread (action_answer, data);
      break;
    case ACTION_INSERT_MENU:
      insert_menu (locale (lc_Locate));
      break;
    case ACTION_MENU_SELECTION:
      stk_thread (locate, data);
      break;
    default:
      break;
  }
}
