/*
 * locserver.c - location info server application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>

u8 PROGMEM t_Position_mime[] = "locinfo";

/* *INDENT-OFF* */

lc_char PROGMEM lc_Locinfoserver[]={
	LC_EN("Locinfo Server")
	LC_END
};

/* *INDENT-ON* */

u8 PROGMEM t_null_sms[] = "";

void locinfo_server (u8 * s)
{
  u8 *tmp = NULL;
  u8 *res;
  u8 *t = NULL;
  u8 *_tp = terminal_profile ();
  u8 *q;
  u8 *ln;
  u8 j;
  u8 tp;

  dbsp (">locinfo\n");

  t = msisdncpy (tpdu_seek (s, T_SMS_AA), MSISDN_SMS, MEM_R);
  if (t == NULL)
    return;

  tmp = malloc (139);
  if (tmp == NULL)
  {
    free (t);
    return;
  }

  q = create_head_a (tmp, s);

  // compose all loc info sms
  *q = 0;
  if (_tp)
  {
    if (_tp[0] >= 4 && _tp[4] & 64)
      *q |= 1;			// B4_7 LOC_NFO
    if (_tp[0] >= 4 && _tp[4] & 128)
      *q |= 2;			// B4_8 NMR
    if (_tp[0] >= 9 && _tp[9] & 4)
      *q |= 4;			// B9_3 BCCH
    if (_tp[0] >= 9 && _tp[9] & 16)
      *q |= 8;			// B9_5 TA
  }
  tp = *q;
  q++;
  ln = q;
  q++;

  if (tp & 0x01)
    // provide_local_info
    // T_LOCATION_INFO
    // 12.19 len=0x07+2
  {
    res = provide_local_info ();
    j = get_tag (res, T_LOCATION_INFO);
    if (j != 0)
    {
      memcpy (q, res + j, res[j + 1] + 2);
      q += res[j + 1] + 2;
    }
  }

  if (tp & 0x02)
  {
    // provide_nmr
    // T_NMR
    // 19.22 len=0x10+2
    // provide_bcch & term_profile
    // T_BCCH_CHANNEL_LIST
    // 12.29 + 1 term_prof
    res = provide_nmr ();
    j = get_tag (res, T_NMR);
    if (j != 0)
    {
      memcpy (q, res + j, res[j + 1] + 2);
      q += res[j + 1] + 2;
    }

    j = get_tag (res, T_BCCH_CHANNEL_LIST);
    if (j != 0)
    {
      memcpy (q, res + j, res[j + 1] + 2);
      q += res[j + 1] + 2;
    }
  }

  if (tp & 0x08)
  {
    // provide_ta
    // T_TIMING_ADVANCE
    // l2.46 len=0x02+2
    res = provide_ta ();
    j = get_tag (res, T_TIMING_ADVANCE);
    if (j != 0)
    {
      memcpy (q, res + j, res[j + 1] + 2);
      q += res[j + 1] + 2;
    }
  }

  *ln = q - ln;
  res =
    send_sms (tmp, q - tmp, t, MSISDN_SMS, tsms_dcs (), tsms_pid (),
	      t_null_sms, NULL);

  free (tmp);
  free (t);
  dbsp ("<locinfo\n");
}

void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_SMS_Q:
      stk_thread (locinfo_server, data);
      break;
    case ACTION_APP_REGISTER:
      reg_sms_tag (locale (lc_Locinfoserver), t_Position_mime, SMS_TYPE_QA);
      break;
    default:
      break;
  }
}
