#include <stdio.h>
#include <unistd.h>

/*
TODO
- cids of same lac must not be on the NEXT_CELL boundaries - offset +- NEXT_CELL

*/

#define TAG_CELL_TEXT		0x01
#define TAG_CELLS_COUNTRY	0x02
#define TAG_CELLS_OPERATOR	0x03
#define TAG_CELLS_DATE    	0x04

typedef struct _Cell_text
{
  char *text;
  unsigned file_pos;
  unsigned text_len;
}
Cell_text;

typedef struct _Cell
{
  unsigned cid;
  unsigned lac;
  Cell_text *text;
}
Cell;

int sort_cells (Cell * a, Cell * b)
{
  if (a->cid < b->cid)
    return -1;
  if (a->cid > b->cid)
    return 1;
  if(a->lac > b->lac)
     return -1;
  if(a->lac < b->lac)
     return 1;
  return 0;
}

void set_bit (unsigned char *buf, unsigned bit, unsigned char val)
{
  unsigned i = bit / 8;
  unsigned char j = bit % 8;

  j = 1 << j;

  if (val == 0)
  {
    j = ~j;
    buf[i] &= j;
  }
  else
  {
    buf[i] |= j;
  }
}

unsigned char get_bit (unsigned char *buf, unsigned bit)
{
  unsigned i = bit / 8;
  unsigned char j = bit % 8;
  unsigned char x = buf[i];

  j = 1 << j;
  x = x & j;
  if (x)
    return 1;
  else
    return 0;
}

unsigned char dcs_8to7 (unsigned char *buf, unsigned char nr_chars)
{
  unsigned bit_len;
  unsigned len_text;
  unsigned x;
  unsigned i, j, k, n;

  len_text = nr_chars;
  for (k = 0; k < nr_chars; k++)
    for (n = 0; n < 7; n++)
    {
      set_bit (buf, k * 7 + n, get_bit (buf, k * 8 + n));
    }

  if (len_text * 7 % 8 != 0)
    len_text = len_text * 7 / 8 + 1;
  else
    len_text = len_text * 7 / 8;

  if (nr_chars * 7 % 8 == 1)
  {
    buf[len_text - 1] &= 0x01;
    buf[len_text - 1] |= '\r' << 1;
  }
  return len_text;
}

unsigned char nodiag1250 (unsigned char x)
{
/*

   CP1250

*/

  switch (x)
  {
//      80              
//      81              
//      82      ???     SINGLE LOW-9 QUOTATION MARK
//      83              
//      84      ???     DOUBLE LOW-9 QUOTATION MARK
//      85      ...     MIDLINE HORIZONTAL ELLIPSIS
//      86      +       DAGGER
//      87      ++      DOUBLE DAGGER
//      88              
//      89      0/00    PER MILLE SIGN
    case 0x8A:			//S(    LATIN CAPITAL LETTER S WITH CARON
      return 'S';
//      8B      <       SINGLE LEFT-POINTING ANGLE QUOTATION MARK
//      8C      OE      LATIN CAPITAL LETTER S WITH ACUTE
    case 0x8D:			//?     LATIN CAPITAL LETTER T WITH CARON
      return 'T';
    case 0x8E:			//Z(    LATIN CAPITAL LETTER Z WITH CARON
      return 'Z';
//      8F      ?       LATIN CAPITAL LETTER Z WITH ACUTE
//      90              
//      91      '       LEFT SINGLE QUOTATION MARK
//      92      '       RIGHT SINGLE QUOTATION MARK
//      93      "       LEFT DOUBLE QUOTATION MARK
//      94      "       RIGHT DOUBLE QUOTATION MARK
//      95      .       BULLET OPERATOR
//      96      -       EN DASH
//      97      --      EM DASH
//      98              
//      99      ^(TM)   TRADE MARK SIGN
    case 0x9A:			//s(    LATIN SMALL LETTER S WITH CARON
      return 's';
//      9B      >       SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
//      9C      oe      LATIN SMALL LETTER S WITH ACUTE
    case 0x9D:			//?     LATIN SMALL LETTER T WITH CARON
      return 't';
    case 0x9E:			//z(    LATIN SMALL LETTER Z WITH CARON
      return 'z';
//      9F      Y"      LATIN SMALL LETTER Z WITH ACUTE
//      A0              NO-BREAK SPACE
//      A1             CARON
//      A2             BREVE
//      A3             LATIN CAPITAL LETTER L WITH STROKE
//      A4             CURRENCY SIGN
//      A5             LATIN CAPITAL LETTER A WITH OGONEK
//      A6             BROKEN BAR
//      A7             SECTION SIGN
//      A8             DIAERESIS
//      A9             COPYRIGHT SIGN
//      AA             LATIN CAPITAL LETTER S WITH CEDILLA
//      AB             LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
//      AC             NOT SIGN
//      AD             SOFT HYPHEN
//      AE             REGISTERED SIGN
//      AF             LATIN CAPITAL LETTER Z WITH DOT ABOVE

//      B0             DEGREE SIGN
//      B1             PLUS-MINUS SIGN
//      B2             OGONEK
//      B3             LATIN SMALL LETTER L WITH STROKE
//      B4             ACUTE ACCENT
//      B5             MICRO SIGN
//      B6             PILCROW SIGN
//      B7             MIDDLE DOT
//      B8             CEDILLA
//      B9             LATIN SMALL LETTER A WITH OGONEK
//      BA             LATIN SMALL LETTER S WITH CEDILLA
//      BB             RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
//      BC             LATIN CAPITAL LETTER Y WITH DIAERESIS
//      BD             DOUBLE ACUTE ACCENT
    case 0xBE:			//     LATIN CAPITAL LETTER L WITH CARON
      return 'l';

//      BF             LATIN SMALL LETTER Z WITH DOT ABOVE
//      C0             LATIN CAPITAL LETTER R WITH ACUTE
    case 0xC1:			//        LATIN CAPITAL LETTER A WITH ACUTE
      return 'A';
//      C2             LATIN CAPITAL LETTER A WITH CIRCUMFLEX
//      C3             LATIN CAPITAL LETTER A WITH BREVE
//      C4             LATIN CAPITAL LETTER A WITH DIAERESIS
//      C5             LATIN CAPITAL LETTER L WITH ACUTE
//      C6             LATIN CAPITAL LETTER C WITH ACUTE
//      C7             LATIN CAPITAL LETTER C WITH CEDILLA
    case 0xC8:			//        LATIN CAPITAL LETTER C WITH CARON
      return 'C';
    case 0xC9:			//        LATIN CAPITAL LETTER E WITH ACUTE
      return 'E';
//      CA             LATIN CAPITAL LETTER E WITH OGONEK
//      CB             LATIN CAPITAL LETTER E WITH DIAERESIS
    case 0xCC:			//        LATIN CAPITAL LETTER E WITH CARON
      return 'E';
    case 0xCD:			//        LATIN CAPITAL LETTER I WITH ACUTE
      return 'I';
//      CE             LATIN CAPITAL LETTER I WITH CIRCUMFLEX
    case 0xCF:			//        LATIN CAPITAL LETTER D WITH CARON
      return 'D';
//      D0             LATIN CAPITAL LETTER D WITH STROKE
//      D1             LATIN CAPITAL LETTER N WITH ACUTE
    case 0xD2:			//        LATIN CAPITAL LETTER N WITH CARON
      return 'N';
    case 0xD3:			//        LATIN CAPITAL LETTER O WITH ACUTE
      return 'O';
//      D4             LATIN CAPITAL LETTER O WITH CIRCUMFLEX
//      D5             LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
//      D6             LATIN CAPITAL LETTER O WITH DIAERESIS
//      D7             MULTIPLICATION SIGN
    case 0xD8:			//        LATIN CAPITAL LETTER R WITH CARON
      return 'R';
    case 0xD9:			//     LATIN CAPITAL LETTER U WITH RING ABOVE
    case 0xDA:			//     LATIN CAPITAL LETTER U WITH ACUTE
//      DB             LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
//      DC             LATIN CAPITAL LETTER U WITH DIAERESIS
      return 'U';
    case 0xDD:			//        LATIN CAPITAL LETTER Y WITH ACUTE
      return 'Y';
    case 0xDE:			//     LATIN CAPITAL LETTER T WITH CEDILLA
      return 't';
    case 0xDF:			//        LATIN SMALL LETTER SHARP S
      return 's';

//      E0             LATIN SMALL LETTER R WITH ACUTE
    case 0xE1:			//        LATIN SMALL LETTER A WITH ACUTE
    case 0xE2:			//        LATIN SMALL LETTER A WITH CIRCUMFLEX
    case 0xE3:			//        LATIN SMALL LETTER A WITH BREVE
    case 0xE4:			//        LATIN SMALL LETTER A WITH DIAERESIS
      return 'a';
    case 0xE5:			//     LATIN SMALL LETTER L WITH ACUTE
      return 'l';
    case 0xE6:			//        LATIN SMALL LETTER C WITH ACUTE
    case 0xE7:			//        LATIN SMALL LETTER C WITH CEDILLA
    case 0xE8:			//        LATIN SMALL LETTER C WITH CARON
      return 'c';
    case 0xE9:			//        LATIN SMALL LETTER E WITH ACUTE
    case 0xEA:			//        LATIN SMALL LETTER E WITH OGONEK
    case 0xEB:			//        LATIN SMALL LETTER E WITH DIAERESIS
    case 0xEC:			//        LATIN SMALL LETTER E WITH CARON
      return 'e';
    case 0xED:			//        LATIN SMALL LETTER I WITH ACUTE
    case 0xEE:			//        LATIN SMALL LETTER I WITH CIRCUMFLEX
      return 'i';
    case 0xEF:			//        LATIN SMALL LETTER D WITH CARON
    case 0xF0:			//        LATIN SMALL LETTER D WITH STROKE
      return 'd';
    case 0xF1:			//        LATIN SMALL LETTER N WITH ACUTE
    case 0xF2:			//        LATIN SMALL LETTER N WITH CARON
      return 'n';
    case 0xF3:			//        LATIN SMALL LETTER O WITH ACUTE
    case 0xF4:			//        LATIN SMALL LETTER O WITH CIRCUMFLEX
    case 0xF5:			//        LATIN SMALL LETTER O WITH DOUBLE ACUTE
    case 0xF6:			//        LATIN SMALL LETTER O WITH DIAERESIS
      return 'o';
//      F7             DIVISION SIGN
    case 0xF8:			//        LATIN SMALL LETTER R WITH CARON
      return 'r';
    case 0xF9:			//        LATIN SMALL LETTER U WITH RING ABOVE
    case 0xFA:			//        LATIN SMALL LETTER U WITH ACUTE
    case 0xFB:			//        LATIN SMALL LETTER U WITH DOUBLE ACUTE
    case 0xFC:			//        LATIN SMALL LETTER U WITH DIAERESIS
      return 'u';
    case 0xFD:			//        LATIN SMALL LETTER Y WITH ACUTE
      return 'y';
    case 0xFE:			//        LATIN SMALL LETTER T WITH CEDILLA
      return 't';
    default:
      return x;
  }
}

void help ()
{
  printf ("Usage: [args] file\n");
  printf ("-h\thelp\n");
  printf ("-d date\n");
  printf ("-c country\n");
  printf ("-o operator\n");
  exit (0);
}

int main (int argc, char **argv)
{
  FILE *f;
  char *t;
  unsigned i, j;

#define LINE_LEN	1000
  char line[LINE_LEN];
  unsigned cellid;
  unsigned lac;
  unsigned nr_cids = 0;
  unsigned nr_texts = 0;
  unsigned nr_cells = 0;

#define NR_CELLS 50000
  Cell_text *texts = malloc (NR_CELLS * sizeof (Cell_text));
  Cell_text *txt;
  Cell *cells = malloc (NR_CELLS * sizeof (Cell));
  unsigned char *res;
  unsigned file_len = 0;
  unsigned file_pos = 0;
  unsigned ver_len = 0;
  char *cells_country = NULL;
  char *cells_date = NULL;
  char *cells_operator = NULL;
  char *file_name = NULL;
  int result;

  do
  {
    result = getopt (argc, argv, "hd:c:o:");
    switch (result)
    {
      case 'd':
	cells_date = optarg;
	break;
      case 'o':
	cells_operator = optarg;
	break;
      case 'c':
	cells_country = optarg;
	break;
      case 'h':
	help ();
	break;
    }
  }
  while (result > 0);

  if (argc - optind != 1)
    help ();

  file_name = argv[optind++];

  f = fopen (file_name, "r");
  if (f == NULL)
  {
    fprintf (stderr, "Can't open file %s\n", f);
    exit (1);
  }

  while (fgets (line, LINE_LEN, f))
  {
    nr_cids++;
    //printf("L: %s\n",s);
    sscanf (line, "%X", &cellid);
    t = strstr (line, ";");
    t++;
    i=sscanf (t, "%d", &lac); if(i==0)lac=0;	//unknown LAC
    t = strstr (t, ";");
    t++;
    for (i = 0; i < strlen (t); i++)
      if (t[i] == '\n')
	t[i] = ' ';
    printf ("CID %04X LAC %d %s\n", cellid, lac, t);

    txt = NULL;
    for (i = 0; i < nr_cells; i++)
    {
      if (cells[i].cid == cellid && cells[i].lac == lac)
      {
	printf ("Duplicate cell %04X %d %s, has %04X %d %s\n", cellid, lac, t,
		cells[i].cid, cells[i].lac, cells[i].text->text);
	t = NULL;
	break;
      }

      if (strcmp (t, cells[i].text->text) == 0)
      {
	printf ("Duplicate text\n");
	txt = cells[i].text;
      }
    }

    if (t)
    {
      cells[nr_cells].cid = cellid;
      cells[nr_cells].lac = lac;
      if (txt == NULL)
      {
	texts[nr_texts].text_len = strlen (t);
	texts[nr_texts].text = malloc (strlen (t) + 1);
	strcpy (texts[nr_texts].text, t);
	cells[nr_cells].text = &texts[nr_texts];
	nr_texts++;
      }
      else
      {
	cells[nr_cells].text = txt;
      }
      nr_cells++;
    }
  }

  fclose (f);

  qsort (cells, nr_cells, sizeof (Cell), sort_cells);

//no dicritics
  for (i = 0; i < nr_texts; i++)
  {
    for (j = 0; j < texts[i].text_len; j++)
      texts[i].text[j] = nodiag1250 (texts[i].text[j]);
  }

//conv ->7b
  for (i = 0; i < nr_texts; i++)
  {
#define MAX_LEN_TEXT 120

    if (texts[i].text_len > MAX_LEN_TEXT)
    {
//      printf("text truncated %d->%d %s\n", texts[i].text_len, MAX_LEN_TEXT, texts[i].text);
      texts[i].text_len = MAX_LEN_TEXT;
    }

    texts[i].text_len = dcs_8to7 (texts[i].text, texts[i].text_len);
  }

#define CELL_VAL_STEP 80

  ver_len = 1;			//head len

  if (cells_country)
    ver_len += 1 + 1 + strlen (cells_country);
  if (cells_date)
    ver_len += 1 + 1 + strlen (cells_date);
  if (cells_operator)
    ver_len += 1 + 1 + strlen (cells_operator);

  file_len = ver_len + nr_cells * (2 + 2 + 4) + nr_cells / CELL_VAL_STEP * 3 + 3;	//index

  for (i = 0; i < nr_texts; i++)
  {
    texts[i].file_pos = file_len;
    file_len += 1 + 1 + 1 + texts[i].text_len;	//overall+tag_text+text_len+text
  }

  res = malloc (file_len);

// dump ver

  file_pos = 0;

  res[file_pos++] = ver_len - 1;

  if (cells_country)
  {
    res[file_pos++] = TAG_CELLS_COUNTRY;
    res[file_pos++] = strlen (cells_country);
    for (i = 0; i < strlen (cells_country); i++)
      res[file_pos++] = cells_country[i];
  }

  if (cells_date)
  {
    res[file_pos++] = TAG_CELLS_DATE;
    res[file_pos++] = strlen (cells_date);
    for (i = 0; i < strlen (cells_date); i++)
      res[file_pos++] = cells_date[i];
  }

  if (cells_operator)
  {
    res[file_pos++] = TAG_CELLS_OPERATOR;
    res[file_pos++] = strlen (cells_operator);
    for (i = 0; i < strlen (cells_operator); i++)
      res[file_pos++] = cells_operator[i];
  }

  for (i = 0; i < nr_cells; i++)
  {
    if (i % CELL_VAL_STEP == 0)
    {
      unsigned next = 0;
      unsigned len = CELL_VAL_STEP;

      if (i + CELL_VAL_STEP < nr_cells)
	next = cells[i + CELL_VAL_STEP].cid;
      else
	len = nr_cells - i;

      printf ("NEXT_CELL STAMP %04X %d %d\n", next, len, i);

      res[file_pos++] = next;
      res[file_pos++] = next >> 8;
      res[file_pos++] = len;
    }

    printf ("CELL %04X LAC %04X %08X\n", cells[i].cid, cells[i].lac, cells[i].text->file_pos);
    res[file_pos++] = cells[i].cid;
    res[file_pos++] = cells[i].cid >> 8;
    res[file_pos++] = cells[i].lac;
    res[file_pos++] = cells[i].lac >> 8;

    res[file_pos++] = cells[i].text->file_pos;
    res[file_pos++] = cells[i].text->file_pos >> 8;
    res[file_pos++] = cells[i].text->file_pos >> 16;
    res[file_pos++] = cells[i].text->file_pos >> 24;
  }

  for (i = 0; i < nr_texts; i++)
  {
    res[texts[i].file_pos] = 1 + 1 + texts[i].text_len;
    res[texts[i].file_pos + 1] = TAG_CELL_TEXT;
    res[texts[i].file_pos + 2] = texts[i].text_len;
    printf ("%08X: %02X %02X %02X\t ", texts[i].file_pos,
	    res[texts[i].file_pos], res[texts[i].file_pos + 1],
	    res[texts[i].file_pos + 2]);
    for (j = 0; j < texts[i].text_len; j++)
    {
      res[texts[i].file_pos + 3 + j] = texts[i].text[j];
      printf ("%02X ", res[texts[i].file_pos + 3 + j]);
    }
    printf ("\n");
  }

//file_len=ver_len+nr_cells*(2+4)+nr_cells/CELL_VAL_STEP*3+3;

  sprintf (line, "%s.cel", file_name);
  f = fopen (line, "w+");
  for (i = 0; i < file_len; i++)
    fputc (res[i], f);
  fclose (f);

  printf ("NR_CIDS %d\n", nr_cids);
  printf ("NR_CELLS %d\n", nr_cells);
  printf ("NR_TEXTS %d\n", nr_texts);
  printf ("FILE_LEN %d\n", file_len);
}
